# -*- coding: utf-8 -*-
"""Implement laposteWS."""
import requests
import email.parser
from lxml import objectify, etree
from jinja2 import Environment, PackageLoader
from roulier.transport import Transport
from roulier.ws_tools import remove_empty_tags, get_parts
from roulier.exception import CarrierError
import logging

log = logging.getLogger(__name__)


class LaposteTransport(Transport):
    """Implement Laposte WS communication."""

    LAPOSTE_WS = "https://ws.colissimo.fr/sls-ws/SlsServiceWS"

    def send(self, payload):
        """Call this function.

        Args:
            payload.body: XML in a string
            payload.header : auth
        Return:
            {
                response: (Requests.response)
                body: XML response (without soap)
                parts: dict of attachments
            }
        """
        body = payload['body']
        headers = payload['headers']
        soap_message = self.soap_wrap(body, headers)
        log.debug(soap_message)
        response = self.send_request(soap_message)
        return self.handle_response(response)

    def soap_wrap(self, body, headers):
        """Wrap body in a soap:Enveloppe."""
        env = Environment(
            loader=PackageLoader('roulier', '/carriers/laposte/templates'),
            extensions=['jinja2.ext.with_'])

        template = env.get_template("laposte_soap.xml")
        body_stripped = remove_empty_tags(body)
        data = template.render(body=body_stripped)
        return data.encode('utf8')

    def send_request(self, body):
        """Send body to laposte WS."""
        return requests.post(
            self.LAPOSTE_WS,
            headers={'content-type': 'text/xml;charset=UTF-8'},
            data=body)

    def handle_500(self, response):
        """Handle reponse in case of ERROR 500 type."""
        log.warning('Laposte error 500')
        obj = objectify.fromstring(response.text)
        errors = [{
            "id": obj.xpath('//faultcode')[0],
            "message": obj.xpath('//faultstring')[0],
        }]
        raise CarrierError(response, errors)

    def handle_200(self, response):
        """
        Handle response type 200 (success).

        It still can be a success or a failure.
        """
        def raise_on_error(response_xml):
            xml = objectify.fromstring(response_xml)
            messages = xml.xpath('//messages')
            errors = [
                {
                    'id': message.id,
                    'message': unicode(message.messageContent),
                }
                for message in messages if message.type == "ERROR"
            ]
            if len(errors) > 0:
                raise CarrierError(response, errors)

        def extract_xml(response):
            """Because the answer is mixedpart we need to extract."""
            content_type = response.headers['Content-Type']
            boundary = content_type.split('boundary="')[1].split('";')[0]
            start = content_type.split('start="')[1].split('";')[0]

            between_boundaries = response.text.split("--%s" % boundary)[1]
            after_start = between_boundaries.split(start)[1]
            clean_xml = after_start.strip()  # = trim()
            return clean_xml

        def extract_body(response_xml):
            """Remove soap wrapper."""
            xml = objectify.fromstring(response_xml)
            payload_xml = xml.Body.getchildren()[0]
            return etree.tostring(payload_xml)

        response_xml = extract_xml(response)
        raise_on_error(response_xml)
        return {
            'body': extract_body(response_xml),
            'parts': get_parts(response),
            'response': response,
        }

    def handle_response(self, response):
        """Handle response of webservice."""
        if response.status_code == 200:
            return self.handle_200(response)
        elif response.status_code == 500:
            return self.handle_500(response)  # will raise
        else:
            raise CarrierError(response, [{
                'id': None,
                'message': "Unexpected status code from server",
            }])

    


# -*- coding: utf-8 -*-
"""Implementation of Geodis Ws Api."""
from roulier.api import Api


class GeodisApiFindLocaliteWs(Api):

    def _service(self):
        schema = {}
        # 'description': 'Use test Ws'
        schema['is_test'] = {'type': 'boolean', 'default': True}
        return schema

    def _address(self):
        schema = super(GeodisApiFindLocaliteWs, self)._address()
        schema['country'].update({'required': True, 'empty': False})
        return {
            'country': schema['country'],
            'zip': schema['zip'],
            'city': schema['city'],
        }

    def _auth(self):
        schema = super(GeodisApiFindLocaliteWs, self)._auth()
        schema['login'].update({'required': True, 'empty': False})
        schema['password']['required'] = False
        return schema

    def _schemas(self):
        return {
            'auth': self._auth(),
            'to_address': self._address(),
            'service': self._service(),
        }



# -*- coding: utf-8 -*-
"""Implementation for Geodis."""

from .geodis_encoder_edi import GeodisEncoderEdi
from .geodis_encoder_ws import GeodisEncoderWs
from .geodis_encoder_rest_ws import GeodisEncoderRestWs
from .geodis_decoder_ws import GeodisDecoderWs
from .geodis_decoder_rest_ws import GeodisDecoderRestWs
from .geodis_transport_ws import GeodisTransportWs
from .geodis_transport_edi import GeodisTransportEdi
from .geodis_transport_rest_ws import GeodisTransportRestWs

from roulier.carrier import Carrier
from roulier.exception import InvalidAction

# Specifications:
# rest
#    usage: tracking
#    document: GEODIS - GUIDE TECHNIQUE SIC - Nouveau service Zoom - v1.1.docx
#    date: 2018
# edi:
#    usage: deposit slip
#    document: IFCSUM_D96A_1026_Client_Construit_FR.doc
#    date: 2012
# xml:
#    usage: label and findLocalite
#    document: GEODIS_Nouvelle eětiquette_GEOLABEL_v FR_V1.9.1.pdf
#    date: 2016


class Geodis(Carrier):
    """Implementation for Geodis."""

    def api(self, action='label'):
        """Expose how to communicate with Geodis."""
        try:
            method = self.ACTIONS[action]
        except KeyError:
            raise InvalidAction("Action not supported")
        return method(self, None, api=True)

    def get(self, data, action):
        """."""
        try:
            method = self.ACTIONS[action]
        except KeyError:
            raise InvalidAction("Action not supported")

        return method(self, data)

    def get_edi(self, data, api=False):
        encoder = GeodisEncoderEdi()
        transport = GeodisTransportEdi()
        if api:
            return encoder.api()
        arr = encoder.encode(data)
        return transport.send(arr)

    def get_label(self, data, api=False):
        """Genereate a demandeImpressionEtiquette."""
        return self._get_ws(data, api, 'demandeImpressionEtiquette')

    def address_validator(self, data, api=False):
        return self._get_ws(data, api, 'findLocalite')

    def get_tracking(self, data, api=False):
        return self._get_rest_ws(data, api, 'tracking')

    def get_tracking_list(self, data, api=False):
        return self._get_rest_ws(data, api, 'trackingList')

    def _get_ws(self, data, api=False, action=None):
        encoder = GeodisEncoderWs()
        decoder = GeodisDecoderWs()
        transport = GeodisTransportWs()

        if api:
            return encoder.api(action=action)

        request = encoder.encode(data, action)
        response = transport.send(request)
        return decoder.decode(
            response['body'],
            response['parts'],
            request['infos'],
        )

    def _get_rest_ws(self, data, api=False, action=None):
        encoder = GeodisEncoderRestWs()
        decode = GeodisDecoderRestWs()
        transport = GeodisTransportRestWs()
        if api:
            return encoder.api(action=action)
        request = encoder.encode(data, action)
        response = transport.send(request)
        return decode.decode(response, action)

    ACTIONS = {
        'label': get_label,
        'findLocalite': address_validator,
        'demandeImpressionEtiquette': get_label,
        'edi': get_edi,
        'tracking': get_tracking,
        'trackingList': get_tracking_list,
    }



# -*- coding: utf-8 -*-

import json
from collections import OrderedDict

from ..geodis_decoder_rest_ws import GeodisDecoderRestWs
from ..geodis_encoder_rest_ws import GeodisEncoderRestWs
from ..geodis_transport_rest_ws import GeodisTransportRestWs
from ..geodis_api_rest_ws import GeodisApiTrackingListOut


def test_encode():
    encoder = GeodisEncoderRestWs()
    api = encoder.api('trackingList')
    payload = encoder.encode(api, 'trackingList')
    infos = payload['infos']
    assert infos['action'] == 'trackingList'
    assert infos['service'] in infos['url']


def test_encode_api():
    encoder = GeodisEncoderRestWs()
    data = encoder.api('trackingList')

    data['service']['shippingDateStart'] = "2019-07-02"
    data['service']['shippingDateEnd'] = "2019-07-02"
    data['auth']['login'] = 'Akretion'
    data['auth']['password'] = '121221789271'

    payload = encoder.encode(data, 'trackingList')
    body = payload['body']
    assert body['dateDepartFin'] == "2019-07-02"


def test_transport_hash():
    """Ensure we encode the has the same way as the specification

    We create the exact same json as they (ordonned, separators)
    """
    api_key = "d89b703bfe0d440a966ff3d996f5936a"
    login = "QTTCLT"
    timestamp = "1546941256145"
    lang = "fr"
    service = "api/zoomclient/recherche-envois"
    params = OrderedDict([
        ("dateDepart", ""),
        ("dateDepartDebut", "2018-12-09"),
        ("dateDepartFin", "2019-01-08"),
        ("noRecepisse", ""),
        ("reference1", ""),
        ("cabColis", ""),
        ("noSuivi", ""),
        ("codeSa", "084135"),
        ("codeClient", ""),
        ("codeProduit", ""),
        ("typePrestation", "EXP"),
        ("dateLivraison", ""),
        ("refDest", ""),
        ("nomDest", ""),
        ("codePostalDest", ""),
        ("natureMarchandise", ""),
    ])
    body = json.dumps(params, separators=(',', ':'))
    transport = GeodisTransportRestWs()
    hash = transport.get_hash(api_key, login, timestamp, lang, service, body)
    token = transport.get_token(login, timestamp, lang, hash)
    expected = (
        "QTTCLT;1546941256145;fr;" +
        "1b59ef28395469bdd3c823adc2603c469c657ed968e96375b0095307e0460fdf")
    assert token == expected


def test_decoder_visit():
    api = GeodisApiTrackingListOut()
    schema = {
        'str': 'simple_string',
        'int': 'simple_integer',
        'inner': {
            'inner_a': 'inner_a',
            'inner_b': 'inner_b',
        }
    }
    data = {
        'simple_string': 'abcdef',
        'simple_integer': 3,
        'inner_a': 'a',
        'inner_b': 'b',
        'no_lookup': 'should not fail',
    }
    ret = api.visit(data, schema)
    assert ret['str'] == data['simple_string']
    assert ret['int'] == data['simple_integer']
    assert ret['inner']['inner_a'] == data['inner_a']


data = (
    """{"contenu": [{"statutServicesLivraison": "","""
    """ "libelleLongEtat": "Livr\u00e9e", "dateEtat": "2019-07-10","""
    """ "libelleLivraison": "Livraison le", "emissionEqc": null,"""
    """ "libelleStatutServicesLivraison": "", "adresse1Exp": "axxxZ","""
    """ "villeExp": "STRASBOURG", "temperatureMin": null, "codeOption":"""
    """ "RET", "libellePrestation": "Retour/Trans. Fr. - Express","""
    """ "urlPreuveService": "", "adresse2Exp": "wcvwv", "dateEtatFrs":"""
    """ "10/07/2019", "avecMatiereDangereuse": false, "dateDepartFrs":"""
    """ "09/07/2019", "urlImageEnlevementLivraison":"""
    """ "http://github.com/akretion/roulier", "emissionEqa": null,"""
    """ "dateLivraisonFrs": "10/07/2019", "noRecepisse": "12212122","""
    """ "nbErreursNotification": 0, "listServicesLivraison": [],"""
    """ "delaiInstruction": 0, "adresse1Dest": "27 rue Henri Rolland","""
    """ "libellePaysExp": "France","""
    """ "codePostalExp": "69100", "envoiRegroupement": false,"""
    """ "refDest": "", "temperatureMed": null,"""
    """ "dateLimiteInstruction": null, "nomDest": "TdsdsfsfsfsT","""
    """ "avecInstructionDonnee": false, "uniteInstruction": "", "refUniEnl":"""
    """ 1212121221, "libelleEtat": "Livr\u00e9e", "nomExp": "RETOUR","""
    """ "urlSuiviDestinataire": "https://akretion.com", "adresse2Dest": "","""
    """ "emissionPar": null, "codePaysDest": "FR", "nbExcursionsTemp": null,"""
    """ "noRecepRegroupement": "", "envoiRegroupe": false, "reference1":"""
    """ "abcdeefefefe", "dateLivraison": "2019-07-10", "loginDestinataire":"""
    """ "12121212121", "nbPalettes": 0, "refUniExp": 1212212121,"""
    """ "avecAttenteInstruction": false, "codePaysExp": "FR", "codeClient":"""
    """ "01234", "listEnvoisRegroupes": [], "codeProduit": "ENE","""
    """ "codeJustification": "NEM", "villeDest": "LESQUIN", "nbColis": 2,"""
    """ "codePostalDest": "59810", "libellePaysDest": "France","""
    """ "typePrestation": "EXP", "refUniRegroupement": 0, "codeSa":"""
    """ "122121", "noSuivi": "12122121", "dateLimiteInstructionFrs": "","""
    """ "codeSituation": "LIV", "dateDepart": "2019-07-09","""
    """ "temperatureMax": null, "poids": 9.0, "reference2": ""},"""
    """ {"statutServicesLivraison": "", "libelleLongEtat": "Livr\u00e9e","""
    """ "dateEtat": "2019-07-10", "libelleLivraison": "Livraison le","""
    """ "emissionEqc": null, "libelleStatutServicesLivraison": "","""
    """ "adresse1Exp": "Akretion", "villeExp": "VILLEURBANNE","""
    """ "temperatureMin": null, "codeOption": "RET","""
    """ "libellePrestation": "Retour/Trans. Fr. - Express","""
    """ "urlPreuveService": "", "adresse2Exp":"""
    """ "GRAND", "dateEtatFrs": "10/07/2019", "avecMatiereDangereuse":"""
    """ false, "dateDepartFrs": "09/07/2019","""
    """ "urlImageEnlevementLivraison": "http://github.com/akretion/roulier","""
    """ "emissionEqa": null, "dateLivraisonFrs": "10/07/2019", """
    """ "noRecepisse": "1212121ae", "nbErreursNotification": 0,"""
    """ "listServicesLivraison": [], "delaiInstruction": 0,"""
    """ "adresse1Dest": "27 Rue Henri Rolland","""
    """ "libellePaysExp": "France", "codePostalExp": "69100","""
    """ "envoiRegroupement": false, "refDest": "","""
    """ "temperatureMed": null, "dateLimiteInstruction": null, "nomDest":"""
    """ "Akretion", "avecInstructionDonnee": false, "uniteInstruction":"""
    """ "", "refUniEnl": 121212, "libelleEtat": "Livr\u00e9e", "nomExp":"""
    """ "RETOUR POUR AVARIE", "urlSuiviDestinataire": "http://akretion.com","""
    """ "adresse2Dest": "", "emissionPar": null, "codePaysDest": "FR","""
    """ "nbExcursionsTemp": null, "noRecepRegroupement": "","""
    """ "envoiRegroupe": false, "reference1": "Shopinvader","""
    """ "dateLivraison": "2019-07-10", "loginDestinataire": "1279217912121","""
    """ "nbPalettes": 0, "refUniExp": 12321212,"""
    """ "avecAttenteInstruction": false, "codePaysExp": "FR", "codeClient":"""
    """ "1212121", "listEnvoisRegroupes": [], "codeProduit": "ENE","""
    """ "codeJustification": "NEM", "villeDest": "Villeurbanne","""
    """ "nbColis": 2, "codePostalDest": "69100","""
    """ "libellePaysDest": "France","""
    """ "typePrestation": "EXP", "refUniRegroupement": 0, "codeSa": "12121","""
    """ "noSuivi": "12212121", "dateLimiteInstructionFrs": "","""
    """ "codeSituation": "LIV", "dateDepart": "2019-07-09","""
    """ "temperatureMax": null, "poids":"""
    """ 5.0, "reference2": ""}], "codeErreur": null,"""
    """ "ok": true, "texteErreur": null}"""
)

ret_val = [{
    'from_address': {
        'city': u'STRASBOURG',
        'country': u'FR',
        'country_name': u'France',
        'name': u'RETOUR',
        'street1': u'axxxZ',
        'street2': u'wcvwv',
        'zip': u'69100'},
    'parcels': {'weight': 9.0},
    'service': {
        'agencyId': u'122121',
        'customerId': u'01234',
        'option': u'RET',
        'product': u'ENE',
        'reference1': u'abcdeefefefe',
        'reference2': u'',
        'reference3': u'',
        'shippingDate': u'2019-07-09',
        'shippingId': u'12212122'},
    'to_address': {
        'city': u'LESQUIN',
        'country': u'FR',
        'country_name': u'France',
        'name': u'TdsdsfsfsfsT',
        'street1': u'27 rue Henri Rolland',
        'street2': u'',
        'zip': u'59810'},
    'tracking': {
        'estDeliveryDate': u'2019-07-10',
        'proofUrl': u'http://github.com/akretion/roulier',
        'publicUrl': u'https://akretion.com',
        'status': 'DELIVERED',
        'statusDate': u'2019-07-10',
        'statusDetails': u'Livr\xe9e',
        'trackingCode': u'12122121'}
}, {
    'from_address': {
        'city': u'VILLEURBANNE',
        'country': u'FR',
        'country_name': u'France',
        'name': u'RETOUR POUR AVARIE',
        'street1': u'Akretion',
        'street2': u'GRAND',
        'zip': u'69100'},
    'parcels': {'weight': 5.0},
    'service': {
        'agencyId': u'12121',
        'customerId': u'1212121',
        'option': u'RET',
        'product': u'ENE',
        'reference1': u'Shopinvader',
        'reference2': u'',
        'reference3': u'',
        'shippingDate': u'2019-07-09',
        'shippingId': u'1212121ae'},
    'to_address': {
        'city': u'Villeurbanne',
        'country': u'FR',
        'country_name': u'France',
        'name': u'Akretion',
        'street1': u'27 Rue Henri Rolland',
        'street2': u'',
        'zip': u'69100'},
    'tracking': {
        'estDeliveryDate': u'2019-07-10',
        'proofUrl': u'http://github.com/akretion/roulier',
        'publicUrl': u'http://akretion.com',
        'status': 'DELIVERED',
        'statusDate': u'2019-07-10',
        'statusDetails': u'Livr\xe9e',
        'trackingCode': u'12212121'
    }}
]


def test_decode():
    payload = json.loads(data)
    rep = {
        "body": payload.get('contenu', []),
        "parts": [],
        "response": None,
    }
    decode = GeodisDecoderRestWs()
    ret = decode.decode(rep, 'trackingList')

    assert ret == ret_val



# -*- coding: utf-8 -*-
"""Exception classes"""
import logging

log = logging.getLogger(__name__)


class InvalidApiInput(Exception):
    """Bad input.

    Use this class in your application to manage
    exception with api call
    """


class InvalidAction(Exception):
    """Bad action requested.

    Actions are WS method or else
    """


class CarrierError(Exception):
    """Error from WS.

    Use this class in your application to manage
    exception with the carrier WS
    """
    def __init__(self, response, msg=None):
        if msg is None:
            msg = "An error occured with WS"
        super(CarrierError, self).__init__(msg)
        self.response = response
        if self.response:
            log.debug(response.text)



# -*- coding: utf-8 -*-
"""Implement geodisWS."""
from roulier.transport import Transport
from datetime import datetime
import logging

log = logging.getLogger(__name__)


class GeodisTransportEdi(Transport):
    """Implement Geodis EDI"""

    def send(self, payload):
        """Call this function.

        Args:
            payload.body: XML in a string
            payload.header : auth
        Return:
            {
                response: (Requests.response)
                body: XML response (without soap)
                parts: dict of attachments
            }
        """
        body = payload['body']
        headers = payload['headers']
        message = self.transport_wrap(body, headers)
        return self.convert_to_edi(message)

    def transport_wrap(self, body, headers):
        date = datetime.now()
        return [
            ['UNB', ['UNOA', '3'],
                [headers['interchangeSender'], '22'],
                [headers['interchangeRecipient'], '22'],
                [date.strftime('%d%m%y'), date.strftime('%H%M')],
                headers['depositId']
             ],
        ] + body + [['UNZ', '1', headers['depositId']]]

    def convert_to_edi(self, arr):
        def parse_token(token):
            if isinstance(token, list):
                return ":".join([sanitize(tok) for tok in token])
            else:
                return sanitize(token)

        def parse_segment(segment):
            return "%s'" % "+".join([parse_token(token) for token in segment])

        def parse_lines(lines):
            return "\n".join([parse_segment(segment) for segment in lines])

        def sanitize(token):
            # remove accents and replace escapable chars by space
            # because, replacing "'" by "?'"
            # increase length of the token
            # which is limited by 35
            sanitized = (
                token
                .replace("?", " ")
                .replace("'", " ")
                .replace("+", " ")
                .replace(":", " ")
            ).encode('ascii', 'ignore') # cut remaining chars
            return sanitized

        return parse_lines(arr)



# -*- coding: utf-8 -*-
"""Implementation of Geodis Api."""
from roulier.codec import Encoder
from roulier.exception import InvalidApiInput
from .geodis_common_ws import GEODIS_INFOS


class GeodisEncoderRestWs(Encoder):

    def api(self, action):
        api = GEODIS_INFOS[action]['api']()
        return api.api_values()

    def encode(self, api_input, action):
        api = GEODIS_INFOS[action]['api']()
        if not api.validate(api_input):
            raise InvalidApiInput(
                'Input error : %s' % api.errors(api_input))
        data = api.normalize(api_input)

        infos = {
            'url': "%s/%s" % (
                GEODIS_INFOS[action]['endpoint'],
                GEODIS_INFOS[action]['service'],
            ),
            'service': GEODIS_INFOS[action]['service'],
            'action': action,
        }

        return {
            'headers': data['auth'],
            'body': data['service'],
            'infos': infos,
        }



# -*- coding: utf-8 -*-
"""Implementation for dummy."""
from .dummy_encoder import DummyEncoder
from .dummy_decoder import DummyDecoder
from .dummy_transport import DummyTransport
from roulier.carrier import Carrier


class Dummy(Carrier):
    """Implementation for dummy."""

    encoder = DummyEncoder()
    decoder = DummyDecoder()
    ws = DummyTransport()

    def api(self):
        """Expose how to communicate with dummy."""
        return self.encoder.api()

    def get(self, data, action):
        """Run an action with data against Laposte WS."""
        request = self.encoder.encode(data, action)
        response = self.ws.send(request)

        if not response['payload']:
            return response
        return self.decoder.decode(response['payload'])

    # shortcuts
    def get_label(self, data):
        """Generate a label."""
        return self.get(data, 'generateLabel')



# -*- coding: utf-8 -*-
"""Utilities for WS."""
from lxml import etree
from jinja2 import Environment, PackageLoader
from zplgrf import GRF
from PIL import Image
from io import BytesIO
import email.parser
import re
import base64


def remove_empty_tags(xml, ouput_as_string=True):
    """Remove empty tags with xslt transformation.

    param: xml a string or a etree type
    return: unicode string or lxml.etree._XSLTResultTree
    """
    # use Jinja env for getting the path of template file
    # pkg_resouces may be an alternative, but we already
    # have Jinja
    env = Environment(
        loader=PackageLoader('roulier', 'templates'),
        extensions=['jinja2.ext.with_'])
    template = env.get_template("remove_empty_tags.xsl")
    xsl = etree.parse(open(template.filename))
    transform = etree.XSLT(xsl)

    if isinstance(xml, basestring):
        xml = etree.fromstring(xml)
    # else we asume xml is an lxml.etree
    if ouput_as_string:
        return unicode(transform(xml))
    else:
        return transform(xml)


def get_parts(response):
    """Extract parts from headers.

    Params:
        response: a request object
    Returns:
        an array of content-ids
    """
    head_lines = ''
    for k, v in response.raw.getheaders().iteritems():
        head_lines += str(k) + ':' + str(v) + '\n'

    full = head_lines + response.content

    parser = email.parser.Parser()
    decoded_reply = parser.parsestr(full)
    parts = {}
    start = decoded_reply.get_param('start').lstrip('<').rstrip('>')
    i = 0
    for part in decoded_reply.get_payload():
        cid = part.get('content-Id', '').lstrip('<').rstrip('>')
        if (not start or start == cid) and 'start' not in parts:
            parts['start'] = part.get_payload()
        else:
            parts[cid or 'Attachment%d' % i] = part.get_payload()
        i += 1
    return parts


def png_to_zpl(png, rotate):
    u"""Transform a PNG in a suitable format for ZPL.

    .png is converted to GRF format.
    Printing a Portable Network Graphics (PNG) with zpl instructions (like DY)
    is possible but hard (the spec is not really understandable).

    params:
        png : base64 encoded png
        rotate: boolean if true, rotate 90°

    returns:
        a ^GFA instruction with the image.
    """
    def base64_to_png(data):
        return BytesIO(base64.b64decode(data))

    def get_grf(png, rotate):
        u"""Get a GRF from a png.

        params:
            png: BytesIO stream
            rotate: rotate 90° if true
        """
        if rotate:
            png_bytes = BytesIO()
            (Image
                .open(png)
                .rotate(90, expand=True)
                .save(png_bytes, format="PNG"))
            png = png_bytes.getvalue()
            png_bytes.close()

        return GRF.from_image(png, 'DEMO')

    def build_gfa(grf):
        zpl = grf.to_zpl_line(compression=2)

        m = re.search('\~DGR:DEMO.GRF,(\d+),(\d+),(.*)$', zpl)
        size = m.group(1)
        width = m.group(2)
        payload = m.group(3)

        return '''^XA^FO00,00
^GFA,%(size)s,%(size)s,%(width)s,,%(payload)s^XZ''' % (
            {'size': size, 'width': width, 'payload': payload}
        )

    with base64_to_png(png) as png_stream:
        grf = get_grf(png_stream, rotate)
        zpl = build_gfa(grf)
    return zpl



# -*- coding: utf-8 -*-
"""Laposte XML -> Python."""
from lxml import objectify
from roulier.codec import Decoder


class LaposteDecoder(Decoder):
    """Laposte XML -> Python."""

    def decode(self, body, parts, output_format):
        """Laposte XML -> Python."""
        def get_product_inter(msg):
            """Understand a getProductInterResponse."""
            x = {
                "product": msg.product,
                "partnerType": msg.partnerType
            }
            return x

        def generate_label_response(msg):
            """Understand a generateLabelResponse."""
            def get_cid(tag, tree):
                element = tree.find(tag)
                if element is None:
                    return None
                href = element.getchildren()[0].attrib['href']
                # href contains cid:236212...-38932@cfx.apache.org
                return href[len('cid:'):]  # remove prefix

            rep = msg.labelResponse
            cn23_cid = get_cid('cn23', rep)
            label_cid = get_cid('label', rep)

            annexes = []

            if cn23_cid:
                annexes.append({
                    "name": 'cn23',
                    "data": parts.get(cn23_cid),
                    "type": "pdf"
                })

            if rep.find('pdfUrl'):
                annexes.append({
                    "name": "label",
                    "data": rep.find('pdfUrl'),
                    "type": "url"
                })

            return {
                "tracking": {
                    "number": rep.parcelNumber,
                    "partner": rep.find('parcelNumberPartner'),
                },
                "label": {  # main label
                    "name": "label",
                    "data": parts.get(label_cid),
                    "type": output_format
                },
                "parcels": [{
                    "id": 1,
                    "number": rep.parcelNumber,
                    "reference": "",
                    "label": {  # same as main label
                        "name": "label_1",
                        "data": parts.get(label_cid),
                        "type": output_format,
                    }

                }],
                "annexes": annexes
            }

        xml = objectify.fromstring(body)
        tag = xml.tag
        lookup = {
            "{http://sls.ws.coliposte.fr}getProductInterResponse":
                get_product_inter,
            "{http://sls.ws.coliposte.fr}generateLabelResponse":
                generate_label_response
        }
        return lookup[tag](xml.xpath('//return')[0])



# -*- coding: utf-8 -*-
"""Implementation of Geodis Api."""
from datetime import datetime
from roulier.codec import Encoder
from roulier.exception import InvalidApiInput
from .geodis_api_edi import GeodisApiEdi


class GeodisEncoderEdi(Encoder):

    def api(self):
        api = GeodisApiEdi()
        return api.api_values()

    def encode(self, api_input):
        api = GeodisApiEdi()
        if not api.validate(api_input):
            raise InvalidApiInput(
                'Input error : %s' % api.errors(api_input))
        data = api.normalize(api_input)

        return {
            "body": self.encode_agency(
                agency_address=data['agency_address'],
                from_address=data['from_address'],
                shipments=data['shipments'],
                service=data['service']),
            "headers": data['service'],
        }

    def encode_shipment(self, shipment, service, idx):
        packs = shipment['parcels']
        to_address = shipment['to_address']
        # There is company notion in to_address, so if company exists
        # we put it in dest name and person name in contact name.
        # if it does not exist, we just put person name in both dest name
        # and contact name. Same is done in webservice side.
        dest_name = to_address.get('company') or to_address['name']
        contact_name = to_address['name']
        lines = [
            ['CNI', '%s' % idx, shipment['shippingId']],
            ['TSR', '2',  # 4065
                [
                    shipment['product'],
                    '', '',
                    shipment['productOption']  # 7273
                ],
                shipment['productPriority'],  # 4219
                shipment['notifications'],  # 7085 : M, S, P
             ],
            # /*MOA si envoi international */
            # ['MOA', '43', $shipment['cost'], 'EUR'],
            ['FTX', 'DEL', '', '', shipment['reference2']],
            ['TOD', '6', 'PP', shipment['productTOD']],
            ['RFF', ['ADE', service['customerId']]],
            ['RFF', ['ACL', shipment['reference1']]],
            ["NAD", "CN",
                "",
                "",  # C058
                dest_name,
                [to_address['street1'], to_address['street2']],
                to_address['city'],
                "",  # 3164
                to_address['zip'],
                to_address['country']
             ],
             ['CTA', 'IC', ['', contact_name]],
        ]
        if to_address['email']:
            lines.append(['COM', [to_address['email'], 'EM']])
        if to_address['phone']:
            lines.append(['COM', [to_address['phone'], 'TE']])
        j = 0
        for pack in packs:
            j = j + 1
            lines += [
                ['GID', '%s' % j, ['1', 'PC', '21', '6']],
                ['MEA', 'AAE', 'AAD', ['KGM', '%s' % pack['weight']]],
                ['PCI', '18'],
                ['GIN', 'BN', pack['barcode']]
            ]
        return lines

    def encode_agency(self, agency_address, from_address, shipments, service):
        shipment_lines = []
        i = 0
        deposit = service['depositId']
        date = datetime.now()
        for shipment in shipments:
            i = i + 1
            shipment_lines += self.encode_shipment(shipment, service, i)

        lines = [
            ['UNH', deposit, ['IFCSUM', 'D', '96A', 'UN', 'ETT021']],
            ['BGM', '630', deposit],
            ['DTM', ['137', date.strftime('%Y%m%d%H%M'), '203']],
            ['RFF', ['ADE', service['customerId']]],
            ['TDT', '20', '', '30', '31'],
            ["NAD", "CA",
                [agency_address['siret'], "100"],
                "",  # C058
                agency_address['name'],
                [agency_address['street1'], agency_address['street2']],
                agency_address['city'],
                "",  # 3164
                agency_address['zip'],
                agency_address['country']
             ],
            ["NAD", "CZ",
                [from_address['siret'], "100"],
                "",  # C058
                from_address['name'],
                [from_address['street1'], from_address['street2']],
                from_address['city'],
                "",  # 3164
                from_address['zip'],
                from_address['country']
             ],
            ['DOC', '630', deposit],
        ] + shipment_lines
        lines += [
            ['UNT', '%s' % (len(lines) + 1), deposit]
        ]
        return lines



# -*- coding: utf-8 -*-
"""Implementation of Geodis Api."""
from roulier.api import Api, MyValidator
from .geodis_api_ws import GeodisApiWs, GEODIS_ALLOWED_NOTIFICATIONS


class GeodisApiEdi(Api):
    def _service(self):
        schema = {
            "depositId": {
                'type': 'string', 'coerce': 'accents',
                'default': '', 'empty': False, 'required': True},
            "depositDate": {
                'type': 'datetime', 'coerce': 'accents',
                'default': '', 'empty': False, 'required': True},
            "customerId": {
                'type': 'string', 'coerce': 'accents',
                'default': '', 'empty': False, 'required': True},
            "interchangeSender": {
                'type': 'string', 'coerce': 'accents',
                'default': '', 'empty': False, 'required': True},
            "interchangeRecipient": {
                'type': 'string', 'coerce': 'accents',
                'default': '', 'empty': False, 'required': True},
        }
        return schema

    def _address(self):
        schema = super(GeodisApiEdi, self)._address()
        schema['country'].update({
            'required': True, 'empty': False,
            'maxlength': 2})
        schema['zip'].update({'required': True, 'empty': False})
        schema['city'].update({'required': True, 'empty': False})
        for key in schema:
            if schema[key].get('type', '') == 'string':
                schema[key].update({'coerce': 'accents'})
        return schema

    def _from_address(self):
        schema = super(GeodisApiEdi, self)._from_address()
        schema['phone'].update({'required': True, 'empty': False})
        schema['street1']['required'] = True
        schema['siret'] = {
            'type': 'string', 'required': True,
            'default': '', 'empty': False}
        for key in schema:
            if schema[key].get('type', '') == 'string':
                schema[key].update({'coerce': 'accents'})
        return schema

    def _parcel(self):
        weight = GeodisApiWs()._parcel()['weight']
        return {
            'weight': weight,
            # 'description': 'Barcode of the parcel'
            'barcode': {
                'type': 'string', 'empty': False, 'default': '',
                'required': True, 'coerce': 'accents'}
        }

    def _to_address(self):
        schema = GeodisApiWs()._to_address()
        for key in schema:
            if schema[key].get('type', '') == 'string':
                schema[key].update({'coerce': 'accents'})
        return schema

    def _shipments(self):
        v = MyValidator()
        schema = {
            'to_address': {
                'type': 'dict',
                'schema': self._to_address(),
                'default': v.normalized({}, self._to_address())
            },
            'parcels': {
                'type': 'list',
                'schema': self._parcel(),
                'default': [v.normalized({}, self._parcel())]
            },
            'product': {
                'type': 'string', 'default': '', 'empty': False,
                'required': True},
            'productTOD': {
                'type': 'string', 'default': '', 'empty': True,
                'required': False},
            'productPriority': {
                'type': 'string', 'default': '', 'empty': True,
                'required': False,
                # 'description': """4219, 1: express, 3: normal speed"""
                },
            'productOption': {
                'type': 'string', 'default': '', 'empty': True,
                'required': False,
                # 'description': """7273, like RDV, B2C, BRT, AEX, A2P..."""
                },
            'notifications': {
                'type': 'string',
                'default': GEODIS_ALLOWED_NOTIFICATIONS[0],
                'allowed': GEODIS_ALLOWED_NOTIFICATIONS,
                'required': False,
                # 'description': """7085 : Notify recipient by M(ail), S(ms), P(=M+S)"""
                },
            'shippingId': {
                'type': 'string', 'default': '', 'empty': False,
                'required': True},
            'reference1': {'type': 'string', 'default': '', 'empty': True},
            'reference2': {'type': 'string', 'default': '', 'empty': True},
            'reference3': {'type': 'string', 'default': '', 'empty': True},
        }
        return {
            "type": "list",
            "schema": {'type': 'dict', 'schema': schema},
            "default": [v.normalized({}, schema)],
        }

    def _schemas(self):
        schemas = {
            'service': self._service(),
            'shipments': self._shipments(),
            'agency_address': self._from_address(),
            'from_address': self._from_address(),
        }
        return schemas



# -*- coding: utf-8 -*-
"""Implementation of Geodis Ws Api."""
from roulier.api import Api

GEODIS_LABEL_FORMAT = (
    'PDF',
    'HTML',
    'EPL',
    'ZPL',
    'XML',
)
GEODIS_ALLOWED_NOTIFICATIONS = (
    'M',  # Mail
    'S',  # SMS
    'P',  # Mail & SMS
)


class GeodisApiWs(Api):
    def _service(self):
        schema = super(GeodisApiWs, self)._service()
        schema['labelFormat']['allowed'] = GEODIS_LABEL_FORMAT
        schema['labelFormat']['default'] = 'ZPL'
        schema['labelFormat'].update({'required': True, 'empty': False})
        schema['product'].update({'required': True, 'empty': False})
        schema['agencyId'].update({'required': False, 'empty': True})
        schema['customerId'].update({'required': True, 'empty': False})
        schema['shippingId'].update({'required': True, 'empty': False})
        schema['hubId'] = {
            # 'description': 'TEOS : code agence Hub de sortie',
            'default': ''
        }
        schema['is_test'] = {
            'type': 'boolean', 'default': True,
            # 'description': 'Use test Ws'
            }
        schema['option'] = {
            'type': 'string', 'default': False,
            # 'description': """Options (RDW, RDV, ETG, ...)"""
            }
        schema['notification'] = {
            'default': GEODIS_ALLOWED_NOTIFICATIONS[0],
            'allowed': GEODIS_ALLOWED_NOTIFICATIONS}

        return schema

    def _address(self):
        schema = super(GeodisApiWs, self)._address()
        schema['country'].update({'required': True, 'empty': False})
        schema['zip'].update({'required': True, 'empty': False})
        schema['city'].update({'required': True, 'empty': False})
        return schema

    def _from_address(self):
        schema = super(GeodisApiWs, self)._from_address()
        schema['phone'].update({'required': True, 'empty': False})
        schema['street1']['required'] = False
        return schema

    def _auth(self):
        schema = super(GeodisApiWs, self)._auth()
        schema['login'].update({'required': True, 'empty': False})
        schema['password']['required'] = False
        return schema

    def _parcel(self):
        schema = super(GeodisApiWs, self)._parcel()
        schema['volume'] = {
            'type': 'float', 'required': False,
            'empty': True, 'default': 0}
        schema['reference'] = {
            'type': 'string', 'required': False,
            'empty': True, 'default': '',
            # 'description': 'Description of this parcel'
            }
        return schema

    def _parcels(self):
        """Allow multiple parcels."""
        schema = super(GeodisApiWs, self)._parcels()
        del schema['items']
        schema['schema'] = self._parcel()
        return schema



# -*- coding: utf-8 -*-
from . import geodis
from . import geodis_common_ws
from . import geodis_api_ws
from . import geodis_api_find_localite_ws
from . import geodis_api_edi
from . import geodis_api_rest_ws
from . import geodis_encoder_ws
from . import geodis_encoder_edi
from . import geodis_encoder_rest_ws
from . import geodis_decoder_ws
from . import geodis_decoder_rest_ws
from . import geodis_transport_ws
from . import geodis_transport_edi
from . import geodis_transport_rest_ws



# -*- coding: utf-8 -*-
"""Implement geodisWS."""
import requests

from roulier.transport import Transport
from roulier.exception import CarrierError
import json
import logging
import hashlib
import time

log = logging.getLogger(__name__)


class GeodisTransportRestWs(Transport):
    """Implement Geodis Rest WS communication."""

    def get_token(self, id, timestamp, lang, hash):
        params = [id, timestamp, lang, hash]
        return ';'.join(params)

    def get_hash(self, api_key, id, timestamp, lang, service, json_data):
        return hashlib.sha256(
            ";".join([
                api_key, id, timestamp, lang, service,
                json_data
            ]).encode("utf-8")
        ).hexdigest()

    def prepare_data(self, data, login, api_key, service):
        timestamp = "%d" % (time.time() * 1000)
        lang = "fr"
        body = json.dumps(data)
        hash = self.get_hash(api_key, login, timestamp, lang, service, body)
        token = self.get_token(login, timestamp, lang, hash)
        return body, token

    def send(self, payload):
        """Call this function.

        Args:
            payload.body: JSON
            payload.header : auth
            payload.infos: { url: string, xmlns: string}
        Return:
            {
                response: (Requests.response)
                body: XML response (without soap)
                parts: empty dict // compat with WS
            }
        """
        body, token = self.prepare_data(
            payload['body'],
            payload['headers']['login'],
            payload['headers']['password'],
            payload['infos']['service']
        )

        infos = payload['infos']
        infos['token'] = token

        response = self.send_request(body, infos)
        log.info('WS response time %s' % response.elapsed.total_seconds())
        return self.handle_response(response)

    def send_request(self, body, infos):
        """Send body to geodis WS."""
        ws_url = infos['url']
        token = infos['token']
        return requests.post(
            ws_url,
            headers={
                'X-GEODIS-Service': token,
            },
            data=body)

    def handle_500(self, response):
        """Handle reponse in case of ERROR 500 type."""
        # TODO : put a try catch (like wrong server)
        log.warning('Geodis error 500')
        errors = [{
            "id": "",
            "message": "",
        }]
        raise CarrierError(response, errors)

    def handle_true_negative_error(self, response, payload):
        """When servers answer an error with a 200 status code."""
        errors = [{
            "id": payload.get('codeErreur', ''),
            "message": payload.get('texteErreur', '')
        }]
        raise CarrierError(response, errors)

    def handle_200(self, response):
        """Handle response type 200."""
        payload = json.loads(response.text)
        if payload['ok'] is not True:
            self.handle_true_negative_error(response, payload)
        return {
            "body": payload.get('contenu', []),
            "parts": [],
            "response": response,
        }

    def handle_response(self, response):
        """Handle response of webservice."""
        if response.status_code == 500:
            return self.handle_500(response)
        elif response.status_code == 200:
            return self.handle_200(response)
        else:
            raise CarrierError(response, [{
                'id': None,
                'message': "Unexpected status code from server",
            }])



# -*- coding: utf-8 -*-
"""Geodis XML -> Python."""
from roulier.codec import Decoder
from .geodis_api_rest_ws import GeodisApiTrackingListOut

class GeodisDecoderRestWs(Decoder):
    """Geodis XML -> Python."""

    def decode(self, payload, action):
        """Geodis JSON -> Python.
        payload[body] : dict
        payload[request] : Requests obj
        """
        body = payload['body']
        if action == 'trackingList':
            mapping = GeodisApiTrackingListOut()
            formatted = []
            for line in body:
                self.add_tracking_code(line)
                formatted.append(
                    mapping.normalize(line)
                )
        else:
            # NOT implemented
            formatted = body
        return formatted



    def add_tracking_code(self, data):
        # MVL: mise en livraison
        # AAR: en cours acheminement
        # None: en attente prise en charge
        # LIV: livré
        # CFM: conforme
        if data['codeSituation'] == 'LIV':
            state = 'DELIVERED'
        elif data['codeSituation'] == 'SOL':
            state = 'RETURNED'
        elif data['codeSituation'] in ('MLV', 'AAR'):
            state = 'TRANSIT'
        else:
            state = 'UNKNOWN'
        data['status'] = state

# CODES=
# 'UNKNOWN',
# 'DELIVERED',
# 'TRANSIT',
# 'FAILURE',
# 'RETURNED'



# -*- coding: utf-8 -*-
"""Store infos about ws"""
from .geodis_api_ws import GeodisApiWs
from .geodis_api_rest_ws import (
    GeodisApiTracking,
    GeodisApiTrackingList,)
from .geodis_api_find_localite_ws import GeodisApiFindLocaliteWs


GEODIS_INFOS = {
    'demandeImpressionEtiquette': {
        'xmlns': 'http://impression.service.web.etiquette.geodis.com',
        'endpoint': "http://espace.geodis.com/geolabel/services/ImpressionEtiquette",  # nopep8
        'endpoint_test': "http://espace.recette.geodis.com/geolabel/services/ImpressionEtiquette",  # nopep8
        'api': GeodisApiWs,
    },
    'findLocalite': {
        'xmlns': 'http://localite.service.web.etiquette.geodis.com',
        'endpoint': "http://espace.geodis.com/geolabel/services/RechercherLocalite",  # nopep8
        'endpoint_test': "http://espace.recette.geodis.com/geolabel/services/RechercherLocalite",  # nopep8
        'api': GeodisApiFindLocaliteWs,
    },
    'trackingList': {
        'endpoint': 'https://espace-client.geodis.com/services',
        'service': 'api/zoomclient/recherche-envois',
        'api': GeodisApiTrackingList,
    },
    'tracking': {
        'endpoint': 'https://espace-client.geodis.com/services',
        'service': 'api/zoomclient/recherche-envoi',
        'api': GeodisApiTracking,
    }
}



# -*- coding: utf-8 -*-
"""Geodis XML -> Python."""
from roulier.codec import Decoder
from .geodis_common_ws import GEODIS_INFOS


class GeodisDecoderWs(Decoder):
    """Geodis XML -> Python."""

    def decode(self, body, parts, infos):
        """Geodis XML -> Python."""

        def reponse_impression_etiquette(msg, parts):
            output_format = infos["output_format"]
            labels = '^XZ•^XA'.join(parts.split('^XZ\r\n^XA')).split('•')
            labels_idx = iter(range(len(labels)))
            labels_data = iter(labels)
            return {
                "tracking": {
                    "number": body.cabRouting,
                },
                "label": {
                    "name": "label",
                    "data": parts,
                    "type": output_format
                },
                "parcels": [{
                    "codeUmg": getattr(colis, 'codumg', ''),
                    "id": getattr(colis, 'numero', ''),
                    "number": getattr(colis, 'cab', ''),
                    "reference": getattr(colis, 'cabclt', ''),
                    "label": {
                        "name": "label_%s" % labels_idx.next(),
                        "data": labels_data.next(),
                        "type": output_format,
                    }
                } for colis in body.infoColis],
                "annexes": [
                ],
                "extra": {
                    "reseau": getattr(body, 'reseau', ''),
                    "priorite": getattr(body, 'priorite', ''),
                    "codeDirectionel": getattr(body, 'codire', ''),
                    "cabRouting": getattr(body, 'cabRouting', ''),
                }
            }

        def response_find_localite(msg, parts):
            return [{
                "ordre": localite.numOrdre,
                "region": localite.codeRegion,
                "zip": localite.codePostal,
                "city": localite.libelle,
            } for localite in msg.infoLocalite]

        tag = body.tag
        lookup = {
            "{http://impression.service.web.etiquette.geodis.com}reponseImpressionEtiquette":
                reponse_impression_etiquette,
            "{http://localite.service.web.etiquette.geodis.com}findLocaliteResponse":
                response_find_localite,
        }
        return lookup[tag](body, parts)



# -*- coding: utf-8 -*-
"""Implementation for Laposte."""

from .laposte_encoder import LaposteEncoder
from .laposte_decoder import LaposteDecoder
from .laposte_transport import LaposteTransport
from roulier.carrier import Carrier


class Laposte(Carrier):
    """Implementation for Laposte."""

    encoder = LaposteEncoder()
    decoder = LaposteDecoder()
    ws = LaposteTransport()

    def api(self):
        """Expose how to communicate with Laposte."""
        return self.encoder.api()

    def get(self, data, action):
        """Run an action with data against Laposte WS."""
        request = self.encoder.encode(data, action)
        response = self.ws.send(request)
        return self.decoder.decode(
            response['body'],
            response['parts'],
            request['output_format']
        )

    # shortcuts
    def get_label(self, data):
        """Genereate a generateLabelRequest."""
        return self.get(data, 'generateLabelRequest')



# -*- coding: utf-8 -*-
"""Implementation for Dpd."""

from .dpd_decoder import DpdDecoder
from .dpd_encoder import DpdEncoder
from .dpd_transport import DpdTransport
from roulier.carrier import Carrier


class Dpd(Carrier):
    """Implementation for Dpd."""

    encoder = DpdEncoder()
    decoder = DpdDecoder()
    ws = DpdTransport()

    def api(self):
        """Expose how to communicate with a Dpd."""
        return self.encoder.api()

    def get(self, data, action):
        """Run an action with data against Dpd WS."""
        request = self.encoder.encode(data, action)
        response = self.ws.send(request)
        return self.decoder.decode(
            response['body'], request['output_format'])

    # shortcuts
    def get_label(self, data):
        """Genereate a createShipmentWithLabels."""
        return self.get(data, 'createShipmentWithLabels')



# -*- coding: utf-8 -*-
"""Transform input to dummy zpl."""
from jinja2 import Environment, PackageLoader
from roulier.codec import Encoder
from roulier.exception import InvalidApiInput
from .dummy_api import DummyApi

DUMMY_ACTIONS = ('generateLabel')


class DummyEncoder(Encoder):
    """Transform input to dummy zpl."""

    def encode(self, api_input, action):
        """Transform input to dummy zpl."""
        if not (action in DUMMY_ACTIONS):
            raise Exception(
                'action %s not in %s' % (action, ', '.join(DUMMY_ACTIONS)))

        api = DummyApi()
        if not api.validate(api_input):
            raise InvalidApiInput(
                'Input error : %s' % api.errors(api_input))
        data = api.normalize(api_input)

        data['to_address']['dept'] = data['to_address']['zip'][0:2]

        env = Environment(
            loader=PackageLoader('roulier', '/carriers/dummy/templates'),
            extensions=['jinja2.ext.with_'])

        template = env.get_template("dummy_%s.zpl" % action)
        return template.render(
            auth=data['auth'],
            service=data['service'],
            parcel=data['parcel'],
            sender_address=data['from_address'],
            receiver_address=data['to_address'])

    def api(self):
        """Return API we are expecting."""
        api = DummyApi()
        return api.api_values()



# -*- coding: utf-8 -*-
"""Transform input to geodis compatible xml."""
from jinja2 import Environment, PackageLoader
from roulier.codec import Encoder
from roulier.exception import InvalidApiInput
from .geodis_common_ws import GEODIS_INFOS


GEODIS_ACTIONS = GEODIS_INFOS.keys()


class GeodisEncoderWs(Encoder):
    """Transform input to geodis compatible xml."""

    def get_api(self, action):
        """Get Api object based on action.

        raise if action is not known"""
        if not (action in GEODIS_ACTIONS):
            raise InvalidApiInput(
                'action %s not in %s' % (
                    action, ', '.join(GEODIS_INFOS)))

        return GEODIS_INFOS[action]['api']()

    def encode(self, api_input, action):
        """Transform input to geodis compatible xml."""
        api = self.get_api(action)
        if not api.validate(api_input):
            raise InvalidApiInput(
                'Input error : %s' % api.errors(api_input))
        env = Environment(
            loader=PackageLoader('roulier', '/carriers/geodis/templates'),
            extensions=['jinja2.ext.with_', 'jinja2.ext.autoescape'],
            autoescape=True)
        template = env.get_template("geodis_%s.xml" % action)

        data = api.normalize(api_input)
        infos = {
            'xmlns': GEODIS_INFOS[action]['xmlns'],
            'url': (
                data['service']['is_test'] and
                GEODIS_INFOS[action]['endpoint_test'] or
                GEODIS_INFOS[action]['endpoint']),
            'action': action,
        }

        def request_impression_etiquette():
            infos["output_format"] = api_input['service']['labelFormat'],
            data['service']['labelFormat'] = self.lookup_label_format(
                data['service']['labelFormat'])
            data['service']['shippingDate'] = (
                data['service']['shippingDate'].replace('/', ''))
            data['from_address']['departement'] = (
                data['from_address']['zip'][:2])
            return {
                "body": template.render(
                    service=data['service'],
                    parcels=data['parcels'],
                    sender_address=data['from_address'],
                    receiver_address=data['to_address'],
                    xmlns=infos['xmlns'],
                ),
                "headers": data['auth'],
                "infos": infos,
            }

        def request_find_localite():
            return {
                "body": template.render(
                    receiver_address=data['to_address'],
                    xmlns=infos['xmlns'],
                ),
                "headers": data['auth'],
                "infos": infos,
            }

        if action == 'demandeImpressionEtiquette':
            return request_impression_etiquette()
        elif action == 'findLocalite':
            return request_find_localite()

    def api(self, action='demandeImpressionEtiquette'):
        """Provide a dict prefilled with default values."""
        return self.get_api(action).api_values()

    def lookup_label_format(self, label_format="ZPL"):
        """Get a Geodis compatible format of label.

        args:
            label_format: (str) ZPL or PDF
        return
            a value taken in GEODIS_LABEL_FORMAT
        """
        lookup = {
            'ZPL': 'Z',
            'PDF': 'P',
            'HTML': 'H',
            'XML': 'X',
            'EPL2': 'E',
            'Z': 'Z',
            'P': 'P',
            'H': 'H',
            'X': 'X',
            'E': 'E',
        }
        return lookup.get(label_format.upper(), 'Z')



# -*- coding: utf-8 -*-
"""API interface."""
from cerberus import Validator
from unidecode import unidecode


class MyValidator(Validator):
    """Custom validator."""

    def _normalize_coerce_zpl(self, value):
        """Sanitze input for ZPL.

        Remove ZPL ctrl caraters
        Remove accents
        """
        if not isinstance(value, basestring):
            return value

        ctrl_cars = [
            0xFE,  # Tilde ~
            0x5E,  # Caret ^
            0x1E,  # RS (^ substitution)
            0x10,  # DLE (~ substitution)
        ]
        val = unidecode(value)
        for ctrl in ctrl_cars:
            val = val.replace("%c" % ctrl, "")
        return val

    def _normalize_coerce_accents(self, value):
        """Sanitize accents for some WS."""
        if not isinstance(value, basestring):
            return value
        sanitized = (
            value
            # quick and dirty replacement
            # of common accentued chars in french
            # because some ws don't handle well utf8
            .replace(u"é", "e")
            .replace(u"è", "e")
            .replace(u"ë", "e")
            .replace(u"ê", "e")
            .replace(u"ô", "o")
            .replace(u"ï", "i")
            .replace(u"ö", "o")
            .replace(u"à", "a")
            .replace(u"â", "a")
            .replace(u"ç", "c")
            .replace(u"û", "u")
            .replace(u"ù", "u")
            .replace(u"É", "E")
            .replace(u"È", "E")
            .replace(u"Ë", "E")
            .replace(u"Ê", "E")
            .replace(u"Ô", "O")
            .replace(u"Ï", "I")
            .replace(u"Ö", "O")
            .replace(u"À", "A")
            .replace(u"Â", "A")
            .replace(u"Ç", "C")
            .replace(u"Û", "U")
            .replace(u"Ù", "U")
            .replace(u"œ", "oe")
            .replace(u"Œ", "OE")
        ).encode('ascii', 'ignore') # cut remaining chars
        return sanitized


class Api(object):
    """Define expected fields of carriers.

    This class should be overriden by each carrier.
    """

    def __init__(self):
        """."""

    def _validator(self):
        v = MyValidator()
        v.allow_unknown = True
        # v.purge_unknown = True
        return v

    def _address(self):
        return {
            'company': {'type': 'string', 'default': ''},
            'name': {'type': 'string', 'default': '', 'required': True, 'empty': False},
            'street1': {'type': 'string', 'default': ''},
            'street2': {'type': 'string', 'default': ''},
            'country': {'type': 'string', 'default': ''},
            # , 'description': 'ISO 3166-1 alpha-2 '},
            'city': {'type': 'string', 'default': ''},
            'zip': {'type': 'string', 'default': ''},
            'phone': {'type': 'string', 'default': ''},
            'email': {'type': 'string', 'default': ''},
        }

    def _from_address(self):
        address = self._address()
        return address

    def _to_address(self):
        address = self._address()
        address['street1'].update({'required': True, 'empty': False})
        address['country'].update({'required': True, 'empty': False})
        address['city'].update({'required': True, 'empty': False})
        address['zip'].update({'required': True, 'empty': False})
        return address

    def _parcel(self):
        return {
            "weight": {'type': 'float', 'default': '', 'required': True, 'empty': False},
        }
        # 'description': 'Weight in kg',

    def _parcels(self):
        v = MyValidator()
        return {
            'type': 'list',
            'items': [{'schema': self._parcel(), 'type': 'dict'}],  # force len=1!
            'default': [v.normalized({}, self._parcel())]
        }

    def _service(self):
        return {
            "product": {'default': ''},
            "agencyId": {'default': ''},
            "customerId": {'default': ''},
            "shippingId": {'default': ''},
            # 'description': 'When the carrier has the package. Format: YYYY/MM/DD'
            'shippingDate': {'default': '', 'type': 'string', 'required': True, 'empty': False},
            # 'description': 'Additionnal info visible by the client. Example : order number'
            'reference1': {'type': 'string', 'default': ''},
            'reference2': {'type': 'string', 'default': ''},
            'reference3': {'type': 'string', 'default': ''},
            # 'description': 'Format of output (usually pdf or zpl)'
            "labelFormat": {'default': ''},
            # 'description': 'Additionnal instructions for delivery',
            "instructions": {'default': ''},
        }

    def _auth(self):
        return {
            'login': {'type': 'string', 'default': ''},
            'password': {'type': 'string', 'default': ''},
        }

    def _schemas(self):
        return {
            'service': self._service(),
            'auth': self._auth(),
            'parcels': self._parcels(),
            'from_address': self._from_address(),
            'to_address': self._to_address(),
        }

    def api_schema(self):
        """Return the expected schema of the api.

        A validation schema is a mapping, usually a dict.
        Schema keys are the keys allowed in the target dictionary.
        Schema values express the rules that must be matched
        by the corresponding target values.

        See http://docs.python-cerberus.org/en/stable/schemas.html

        """
        v = self._validator()
        schemas = self._schemas()

        def wrap_schema(schema):
            # if schema is a simple dict, wrap it as a dict
            # else this work has already be done
            # like it's a list
            if 'schema' in schema or 'items' in schema:
                return schema
            return {
                'schema': schema,
                'default':
                    schema.get('default') or
                    v.normalized({}, schema),
                'type': schema.get('type', 'dict')
            }

        return {
            s: wrap_schema(schemas[s])
            for s in schemas}

    def api_values(self):
        """Return a dict containing expected keys.

        It's a normalized version of the schema.
        """
        return self.normalize({})

    def errors(self, data):
        """Return validation errors."""
        v = self._validator()
        v.validate(data, self.api_schema())
        return v.errors

    def validate(self, data):
        """Ensure the data are valid.

        returns: bool

        See also errors()
        """
        return self._validator().validate(data, self.api_schema())

    def normalize(self, data):
        """Retrurn a normalized dict based on input.

        See http://docs.python-cerberus.org/en/stable/usage.html
        """
        return self._validator().normalized(data, self.api_schema())



# -*- coding: utf-8 -*-
"""Implementation of Laposte Api."""
from roulier.api import Api

LAPOSTE_LABEL_FORMAT = (
    'ZPL_10x15_203dpi',
    'ZPL_10x15_300dpi',
    'DPL_10x15_203dpi',
    'DPL_10x15_300dpi',
    'PDF_10x15_300dpi',
    'PDF_A4_300dpi',
)


class LaposteApi(Api):
    def _service(self):
        schema = super(LaposteApi, self)._service()
        schema['labelFormat']['allowed'] = (
            list(LAPOSTE_LABEL_FORMAT) + ["PDF", "ZPL", "DPL"])
        schema['labelFormat']['default'] = 'ZPL_10x15_203dpi'
        schema['labelFormat_x'] = {'default': 0}
        schema['labelFormat_y'] = {'default': 0}
        schema['labelFormat'].update({'required': True, 'empty': False})
        schema['product'].update({'required': True, 'empty': False})
        schema['pickupLocationId'] = {
            'default': '',
            # 'description': """Si productCode = A2P, BPR, ACP, CDI, CMT, BDP. "
            # "Identifiant du point de retrait "
            # "(dans le cas d’une livraison Colissimo hors domicile)"""
            }
        schema['totalAmount'] = {
            'default': '',
            # 'description': 'Needed for cn23'
            }
        schema['orderNumber'] = {'default': ''}
        schema['commercialName'] = {
            'default': '',
            # 'description':
            # """Obligatoire pour les produits DOM, DOS, BPR, A2P. "
            # "Nom commercial du chargeur qui sera affiché dans "
            # "les notifications par email aux destinataires des colis"""
            }
        schema['returnTypeChoice'] = {
            'default': '',
            # 'description':
            # """Obligatoire pour certains colis à l’international, "
            # "selon les zones tarifaires applicables. Indique si "
            # "le colis doit être retourné à l’expéditeur en cas de "
            # "non distribution du colis"""
            }
        schema['returnType'] = {
            'default': '',
            # 'description':
            # """Utilisé pour le Colissimo Retour uniquement. "
            # "Définit le mode de transmission de l’étiquette"""
            }
        # schema['reference1'].update({
        #     'description': """Référence expediteur ('Réf client')"""
        # })
        # schema['reference2'].update({
        #     'description': """Référence destinataire ('Réf destinataire')"""
        # })
        return schema

    def _address(self):
        schema = super(LaposteApi, self)._address()
        schema['country'].update({'required': True, 'empty': False})
        schema['zip'].update({'required': True, 'empty': False})
        schema['city'].update({'required': True, 'empty': False})
        schema['street0'] = {
            'required': False, 'empty': True,
            # 'description': """Entrée, bâtiment, immeuble, résidence. """
            #                """Non utilisé pour la Belgique."""
            }
        schema['street2'].update({
            'default': ''
            # , 'description': 'Etage, couloir, escalier, appart.'
            })
        schema['street1'].update({
            'required': True, 'empty': False,
            # 'description': 'Numéro et libellé de voie. Ex : 5 rue du Bellay'
            })
        schema['street3'] = {
            'default': '',
            # 'description': """Lieu dit ou autre mention. """
            #                """Non utilisé pour la Belgique."""
            }
        # 'description': """Code porte 1"""
        schema['door1'] = {'default': ''}
        # 'description': """Code porte 2"""
        schema['door2'] = {'default': ''}
        # 'description': """Interphone"""
        schema['intercom'] = {'default': ''}

        return schema

    def _from_address(self):
        schema = super(LaposteApi, self)._from_address()
        return schema

    def _to_address(self):
        schema = super(LaposteApi, self)._to_address()
        schema['firstName'] = {
            'default': '',
            }
            # 'description': """Prénom. Obligatoire pour So Colissimo"""
        return schema

    def _parcel(self):
        schema = super(LaposteApi, self)._parcel()
        schema['nonMachinable'] = {
            'type': 'boolean', 'default': False,
            # 'description': """Passer à true pour indiquer que le "
            # "format du colis est non standard"""
            }
        schema['instructions'] = {
            'default': '',
            # 'description': """Indications complémentaires pour la livraison"""
            }
        schema['insuranceValue'] = {
            'type': 'number', 'default': 0,
            # 'description': """Valeur assurée. Max= 1500€ "
            # "Passer 1230 pour 12,30€ Cette valeur sera arrondie "
            # "à l’entier le plus proche (Ex : 12 euros si 1232 est envoyé) "
            # "Par défaut, renseigner « 0 » (zéro)"""
            }
        schema['recommendationLevel'] = {
            'default': '',
            # 'description': """Niveau de recommandation "
            # "(cf. III.2) Peut valoir « R1 », ou « R2 », ou « R3 »"""
            }
        schema['cod'] = {
            'type': 'boolean', 'default': False,
            # 'description': """Passer à true si la livraison doit se faire contre remboursement."""
            }
        schema['codAmount'] = {
            'type': 'number', 'default': 0,
            # 'description': """Montant attendu lors de la "
            # "livraison contre remboursement. Par défaut, renseigner "
            # "« 0 » (zéro)"""
            }
        schema['ftd'] = {
            'type': 'boolean', 'default': False,
            # 'description': """Pour les envois vers "
            # "l’Outre-Mer uniquement Indique si le colis est franc de taxes "
            # "et de droits."""
            }
        return schema

    def _auth(self):
        schema = super(LaposteApi, self)._auth()
        schema['login'].update({'required': True, 'empty': False})
        schema['password']['required'] = False
        return schema

    def _customs(self):
        schema = {
            'category': {'default': ''},
            'articles': {
                'type': 'list',
                'schema': {
                    'quantity': {'default': ''},
                    'weight': {'type': 'float', 'default': 0.0},
                    'description': {'default': ''},
                    'hs': {'default': ''},
                    'value': {'default': ''},
                    'originCountry': {'default': ''}
                },
                'default': []
            }
        }
        return schema

    def _schemas(self):
        schemas = super(LaposteApi, self)._schemas()
        schemas['customs'] = self._customs()
        return schemas



# -*- coding: utf-8 -*-
"""Implement geodisWS."""
import requests
from lxml import objectify
from jinja2 import Environment, PackageLoader
from roulier.transport import Transport
from roulier.ws_tools import remove_empty_tags, get_parts
from roulier.exception import CarrierError
import logging

log = logging.getLogger(__name__)


class GeodisTransportWs(Transport):
    """Implement Geodis WS communication."""

    def send(self, payload):
        """Call this function.

        Args:
            payload.body: XML in a string
            payload.header : auth
            payload.infos: { url: string, xmlns: string}
        Return:
            {
                response: (Requests.response)
                body: XML response (without soap)
                parts: dict of attachments
            }
        """
        body = payload['body']
        headers = payload['headers']
        infos = payload['infos']
        soap_message = self.soap_wrap(body, headers, infos)
        response = self.send_request(soap_message, infos)
        log.info('WS response time %s' % response.elapsed.total_seconds())
        return self.handle_response(response)

    def soap_wrap(self, body, auth, infos):
        """Wrap body in a soap:Enveloppe."""
        env = Environment(
            loader=PackageLoader('roulier', '/carriers/geodis/templates'),
            extensions=['jinja2.ext.with_'])

        template = env.get_template("geodis_soap.xml")
        body_stripped = remove_empty_tags(body)
        header_template = env.get_template("geodis_header.xml")
        header_xml = header_template.render(auth=auth, xmlns=infos['xmlns'])
        data = template.render(
            body=body_stripped, header=header_xml, xmlns=infos['xmlns'])
        return data.encode('utf8')

    def send_request(self, body, infos):
        """Send body to geodis WS."""
        ws_url = infos['url']
        return requests.post(
            ws_url,
            headers={
                'content-type': 'text/xml',
                'SOAPAction': '<SOAP Action>'
            },
            data=body)

    def handle_500(self, response):
        """Handle reponse in case of ERROR 500 type."""
        # TODO : put a try catch (like wrong server)
        log.warning('Geodis error 500')
        xml = get_parts(response)['start']
        obj = objectify.fromstring(xml)
        message = obj.xpath("//*[local-name() = 'message']")
        id_message = None
        if len(message) > 0:
            message = message[0] or obj.xpath('//faultstring')[0]
            id_message = (
                obj.xpath("//*[local-name() = 'code']") and
                obj.xpath("//*[local-name() = 'code']")[0] or '')
        errors = [{
            "id": id_message,
            "message": message,
        }]
        raise CarrierError(response, errors)

    def handle_200(self, response):
        """Handle response type 200."""
        parts = get_parts(response)
        xml = parts['start']

        def extract_soap(response_xml):
            obj = objectify.fromstring(response_xml)
            return obj.Body.getchildren()[0]

        payload = extract_soap(xml)
        try:
            # TODO : may be extract this elsewere
            # rechercheLocalite has no attachment
            attachement_cid = payload.codeAttachement.text[len('cid:'):]
            attachement = parts[attachement_cid]
        except AttributeError:
            attachement = None

        return {
            "body": payload,
            "parts": attachement,
            "response": response,
        }

    def handle_response(self, response):
        """Handle response of webservice."""
        if response.status_code == 500:
            return self.handle_500(response)
        elif response.status_code == 200:
            return self.handle_200(response)
        else:
            raise CarrierError(response, [{
                'id': None,
                'message': "Unexpected status code from server",
            }])



# -*- coding: utf-8 -*-
"""Transform to/from a carrier specific format."""
import abc


class Encoder(object):
    """Transform to a carrier specific format.

    For preparing (encode) a request to a carrier for stuff like a label
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def encode(self, input):
        """Transform an oject to carrier specific like XML or EDI.

        Args:
            input: python dict
        Return:
            Specific representation of a request of the carrier
        """
        return


class Decoder(object):
    """Transform from a carrier specific format to dict."""

    @abc.abstractmethod
    def decode(self, answer):
        """Transform a specific representation to python dict.

        Args:
            answer: specific representation of an response of the carrier
        Return:
            Python dict
        """
        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

version = open('VERSION').read().strip()
download_url = "https://github.com/akretion/roulier/archive/%s.zip" % version
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="roulier",
    version=version,
    packages=find_packages(),
    install_requires=[
        'lxml', 'Jinja2', 'requests', 'cerberus', 'zplgrf',
        'unidecode'],
    author="Hparfr <https://github.com/hparfr>",
    author_email="roulier@hpar.fr",
    description="Label parcels without pain",
    include_package_data=True,
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_data={'roulier': ['*.xml', '*.xsl', '*.zpl']},
    url="https://github.com/akretion/roulier",
    download_url=download_url,
    keywords=['carrier', 'logistics', 'delivery'],
)



# -*- coding: utf-8 -*-
"""Dpd XML -> Python."""
from lxml import objectify
from roulier.codec import Decoder
from roulier import ws_tools as tools
import base64


class DpdDecoder(Decoder):
    """Dpd XML -> Python."""

    def decode(self, body, output_format):
        """Dpd XML -> Python."""
        def create_shipment_with_labels(msg):
            """Understand a CreateShipmentWithLabelsResponse."""
            shipments, labels = (
                msg.CreateShipmentWithLabelsResult.getchildren()
            )
            shipment = shipments.getchildren()[0]
            label, attachment = labels.getchildren()
            label_data = self.handle_zpl(label.label.text, output_format)
            # .text because we want str instead of objectify.StringElement
            summary_data = base64.b64decode(attachment.label.text)
            summary_format = output_format == 'ZPL' and 'png' or output_format
            x = {
                "tracking": {
                    'number': shipment.barcode.text,
                },
                "parcels": [{
                    "id": 1,
                    "reference": "",
                    "number": shipment.parcelnumber.text,
                    "label": {  # same as main label
                        "data": label_data,
                        "name": "label 1",
                        "type": output_format,
                    }
                }],
                "label": {  # main label
                    "data": label_data,
                    "name": "label",
                    "type": output_format,
                },
                "annexes": [{
                    "data": summary_data,
                    "name": "Summary",
                    "type": summary_format
                }]
            }
            return x

        xml = objectify.fromstring(body)
        tag = xml.tag
        lookup = {
            "{http://www.cargonet.software}CreateShipmentWithLabelsResponse":
                create_shipment_with_labels,
        }
        return lookup[tag](xml)

    def handle_zpl(self, png, label_format):
        """Convert a png in zpl.

        if labelFormat was asked as ZPL, WS returns a png
        This function rotate it and convert it an suitable zpl format
        """
        if label_format == 'ZPL':
            return tools.png_to_zpl(png, True)
        else:
            return base64.b64decode(png)



# -*- coding: utf-8 -*-
"""Implementation of Geodis Api."""
from roulier.api import Api


class GeodisApiRestWs(Api):

    def _schemas(self):
        return {
            'service': self._service(),
            'auth': self._auth(),
        }

    def normalize(self, data):
        externalApi = super(GeodisApiRestWs, self)
        internalApi = self._interal_api()
        step1 = externalApi.normalize(data)
        step2 = internalApi.normalize(step1)
        return step2

    def api_values(self):
        """Return a dict containing expected keys.

        It's a normalized version of the schema.
        only internal api
        """
        return self._validator().normalized({}, self.api_schema())

    def _interal_api(self):
        pass


class GeodisMappingIn(Api):
    """Internal API"""

    def flatten(self, data, out):
        for (key, val) in data.items():
            if isinstance(val, dict):
                self.flatten(val, out)
            else:
                out[key] = val

    def normalize(self, data):
        without_auth = {
            key: val
            for (key, val) in data.items()
            if key != 'auth'
        }
        flat = {
            'auth': data['auth'],
            'service': {}
        }
        self.flatten(without_auth, flat['service'])
        normalized = super(GeodisMappingIn, self).normalize(flat)
        return normalized


class GeodisApiTrackingListMapping(GeodisMappingIn):
    """Internal API

    Used to rename fields."""
    def _schemas(self):
        return {"service": {
            "shippingDate": {"rename": "dateDepart"},
            "shippingDateStart": {"rename": "dateDepartDebut"},
            "shippingDateEnd": {"rename": "dateDepartFin"},
            "agencyId": {"rename": "codeSa"},
            "customerId": {"rename": "codeClient"},
            "reference1": {"rename": "reference1"},
            "reference2": {"rename": "refDest"},
            "name": {"rename": "nomDest"},
            "zip": {"rename": "codePostalDest"},
            'estDeliveryDate': {"rename": 'dateLivraison'},
            "shippingId": {"rename": "noRecepisse"},
            "barcode": {"rename": "cabColis"},
            "trackingId": {"rename": "noSuivi"},
        }}


class GeodisApiTrackingMapping(GeodisMappingIn):
    """Internal API

    Used to rename fields."""
    def _schemas(self):
        return {"service": {
            "trackingId": {"rename": "noSuivi"},
        }}


class GeodisApiTracking(GeodisApiRestWs):

    def _service(self):
        schema = {
            "refUniExp": {"type": "string", "default": "", "empty": True},
            "trackingId": {"type": "string", "default": "", "empty": True},
        }
        return schema

    def _interal_api(self):
        return GeodisApiTrackingMapping()


class GeodisApiTrackingList(GeodisApiRestWs):
    def _service(self):
        return {
            "shippingDate": {"type": "string", "default": "", "empty": True},
            "shippingDateStart": {"type": "string", "default": "", "empty": True},
            "shippingDateEnd": {"type": "string", "default": "", "empty": True},
            "agencyId": {"type": "string", "default": "", "empty": True},
            "customerId": {"type": "string", "default": "", "empty": True},
            "reference1": {"type": "string", "default": "", "empty": True},
            "reference2": {"type": "string", "default": "", "empty": True},
        }

    def _tracking(self):
        return {
            "estDeliveryDate": {"type": "string", "default": "", "empty": True},
            "shippingId": {"type": "string", "default": "", "empty": True},
            "barcode": {"type": "string", "default": "", "empty": True},
            "trackingId": {"type": "string", "default": "", "empty": True},
        }

    def _to_address(self):
        return {
            "name": {"type": "string", "default": "", "empty": True},
            "zip": {"type": "string", "default": "", "empty": True},
        }

    def _schemas(self):
        schema = super(GeodisApiTrackingList, self)._schemas()
        schema['tracking'] = self._tracking()
        schema['to_address'] = self._to_address()
        return schema

    def _interal_api(self):
        return GeodisApiTrackingListMapping()


class GeodisMappingOut(Api):

    def normalize(self, data):
        schema = self.schema()
        # self.add_tracking_code(data)
        return self.visit(data, schema)

    def visit(self, data, schema):
        out = {}
        for (key, val) in schema.items():
            if isinstance(val, dict):
                out[key] = self.visit(data, val)
            else:
                out[key] = data[val]
        return out


class GeodisApiTrackingListOut(GeodisMappingOut):

    def to_address(self):
        return {
            'street1': 'adresse1Dest',
            'street2': 'adresse2Dest',
            'country': 'codePaysDest',
            'zip': 'codePostalDest',
            'country_name': 'libellePaysDest',
            'name': 'nomDest',
            'city': 'villeDest',
        }

    def from_address(self):
        return {
            'street1': 'adresse1Exp',
            'street2': 'adresse2Exp',
            'country': 'codePaysExp',
            'zip': 'codePostalExp',
            'country_name': 'libellePaysExp',
            'name': 'nomExp',
            'city': 'villeExp',
        }

    def parcels(self):
        return {
            'weight': 'poids',
        }

    def service(self):
        return {
            "product": 'codeProduit',
            "agencyId": 'codeSa',
            "customerId": 'codeClient',
            "shippingId": 'noRecepisse',
            'shippingDate': 'dateDepart',
            'reference1': 'reference1',
            'reference2': 'reference2',
            'reference3': 'refDest',
            'option': 'codeOption',
        }

    def tracking(self):
        return {
            'statusDate': 'dateEtat',
            'estDeliveryDate': 'dateLivraison',
            'status': 'status',
            'statusDetails': 'libelleLongEtat',
            'trackingCode': 'noSuivi',
            'publicUrl': 'urlSuiviDestinataire',
            'proofUrl': 'urlImageEnlevementLivraison',
        }

    def schema(self):
        return {
            'parcels': self.parcels(),
            'service': self.service(),
            'from_address': self.from_address(),
            'to_address': self.to_address(),
            'tracking': self.tracking(),
        }



# -*- coding: utf-8 -*-
"""Implementation of Dpd Api."""
from roulier.api import Api

DPD_LABEL_FORMAT = (
    'PDF',
    'PNG',
    'ZPL',
)

DPD_ALLOWED_NOTIFICATIONS = (
    'No',
    'Predict',
    'AutomaticSMS',
    'AutomaticMail',
)
DPD_PRODUCTS = (
    'DPD_Classic',
    'DPD_Relais',
    'DPD_Predict',
    # 'DPD Retour en Relais',  # Not implemented yet
)


class DpdApi(Api):
    def _service(self):
        schema = super(DpdApi, self)._service()
        schema['labelFormat']['allowed'] = DPD_LABEL_FORMAT
        schema['labelFormat']['default'] = 'ZPL'
        schema['labelFormat'].update({'required': True, 'empty': False})
        schema['agencyId'].update({
            'required': True, 'empty': False,
            # 'description': 'Agency code int(3)'
            })
        schema['customerCountry'] = {
            'required': True, 'empty': False,
            # 'description': 'Customer country code (France = 250) int(3)'
            }
        schema['customerId'].update({
            'required': True, 'empty': False,
            # 'description': 'Customer number int(6)'
            })
        schema['shippingDate'].update({'required': False, 'empty': True})

        # mettre ça ensemble ?
        schema['notifications'] = {
            'default': DPD_ALLOWED_NOTIFICATIONS[0],
            'allowed': DPD_ALLOWED_NOTIFICATIONS}
        schema['product'].update({
            'empty': False,
            'required': True,
            'default': DPD_PRODUCTS[0],
            # 'description': 'Type de produit',
            'allowed': DPD_PRODUCTS
        })

        schema['dropOffLocation'] = {
            'default': '',
            'empty': True,
            'required': False,
            # 'description': 'Drop-off Location id (Relais Colis)'
        }

        return schema

    def _address(self):
        schema = super(DpdApi, self)._address()
        # schema['street2']['description'] = (
        #     "N/A for DPD. It will be appended to street1")
        schema['country'].update({'required': True, 'empty': False})
        schema['zip'].update({'required': True, 'empty': False})
        schema['city'].update({'required': True, 'empty': False})
        return schema

    def _to_address(self):
        schema = super(DpdApi, self)._to_address()
        # , 'description': """First name""",
        schema['firstName'] = {'default': '', 'coerce': 'accents'}
        schema['door1'] = {'default': ''}  # 'description': """Door code 1"""
        schema['door2'] = {'default': ''}  # 'description': """Door code 2"""}
        schema['intercom'] = {'default': ''}  # 'description': """Intercom"""}
        for field in ['city', 'company', 'name', 'street1', 'street2']:
            schema[field].update({'coerce': 'accents'})
        return schema

    def _from_address(self):
        schema = super(DpdApi, self)._from_address()
        schema['phone'].update({'required': True, 'empty': False})
        return schema

    def _auth(self):
        schema = super(DpdApi, self)._auth()
        schema['login'].update({'required': True, 'empty': False})
        schema['password']['required'] = False
        return schema



# -*- coding: utf-8 -*-
"""Transform input to dpd compatible xml."""
from jinja2 import Environment, PackageLoader
from roulier.codec import Encoder
from datetime import datetime
from .dpd_api import DpdApi
from roulier.exception import InvalidApiInput
import logging

DPD_ACTIONS = ('createShipmentWithLabels')
log = logging.getLogger(__name__)


class DpdEncoder(Encoder):
    """Transform input to dpd compatible xml."""

    def encode(self, api_input, action):
        """Transform input to dpd compatible xml."""
        if not (action in DPD_ACTIONS):
            raise InvalidApiInput(
                'action %s not in %s' % (action, ', '.join(DPD_ACTIONS)))

        api = DpdApi()
        if not api.validate(api_input):
            raise InvalidApiInput(
                'Input error : %s' % api.errors(api_input))
        data = api.normalize(api_input)

        # add some rules which are hard to implement with
        # cerberus.
        # TODO: add additional schemas for that
        if data['service']['product'] == 'DPD_Predict':
            if len(data['service']['dropOffLocation']) > 0:
                raise InvalidApiInput(
                    "dropOffLocation can't be used with predict")
            if data['service']['notifications'] != 'Predict':
                log.info(
                    'Notification forced to predict because of product')
                data['service']['notifications'] = 'Predict'

        if data['service']['product'] == 'DPD_Classic':
            if len(data['service']['dropOffLocation']) > 0:
                raise InvalidApiInput(
                    "dropOffLocation can't be used with classic")
            if data['service']['notifications'] == 'Predict':
                raise InvalidApiInput(
                    "Predict notifications can't be used with classic")

        if data['service']['product'] == 'DPD_Relais':
            if len(data['service']['dropOffLocation']) < 1:
                raise InvalidApiInput(
                    "dropOffLocation is mandatory for this product")
            if data['service']['notifications'] == 'Predict':
                raise InvalidApiInput(
                    "Predict notifications can't be used with Relais")

        data['service']['shippingDate'] = (
            datetime
            .strptime(data['service']['shippingDate'], '%Y/%M/%d')
            .strftime('%d/%M/%Y')
        )

        def reduce_address(address):
            """Concat some fields.

            Because there is no street2 nor company in DPD api.
            append street2 to street1 and truncate at 70
            append company to name
            """
            address['street1'] = ("%s, %s" % (
                address['street1'], address['street2']))[0:70]
            address['name'] = ("%s, %s" % (
                address['name'], address['company']))[0:35]

        reduce_address(data['to_address'])
        reduce_address(data['from_address'])

        output_format = data['service']['labelFormat']
        if data['service']['labelFormat'] in ('PNG', 'ZPL'):
            # WS doesn't handle zpl yet, we convert it later
            # png is named Default, WTF DPD?
            data['service']['labelFormat'] = 'Default'

        env = Environment(
            loader=PackageLoader('roulier', '/carriers/dpd/templates'),
            extensions=['jinja2.ext.with_', 'jinja2.ext.autoescape'],
            autoescape=True)

        template = env.get_template("dpd_%s.xml" % action)
        return {
            "body": template.render(
                service=data['service'],
                parcel=data['parcels'][0],
                sender_address=data['from_address'],
                receiver_address=data['to_address']),
            "headers": data['auth'],
            "output_format": output_format
        }

    def api(self):
        """Return API we are expecting."""
        return DpdApi().api_values()



# -*- coding: utf-8 -*-
"""Implement dpdWS."""
import requests
from lxml import objectify, etree
from jinja2 import Environment, PackageLoader
from roulier.transport import Transport
from roulier.ws_tools import remove_empty_tags
from roulier.exception import CarrierError

import logging

log = logging.getLogger(__name__)


class DpdTransport(Transport):
    """Implement Dpd WS communication."""

    DPD_WS = "https://e-station.cargonet.software/dpd-eprintwebservice/eprintwebservice.asmx"

    def send(self, payload):
        """Call this function.

        Args:
            payload.body: XML in a string
            payload.header : auth
        Return:
            {
                response: (Requests.response)
                body: XML response (without soap)
            }
        """
        body = payload['body']
        headers = payload['headers']
        soap_message = self.soap_wrap(body, headers)
        log.debug(soap_message)
        response = self.send_request(soap_message)
        log.info('WS response time %s' % response.elapsed.total_seconds())
        return self.handle_response(response)

    def soap_wrap(self, body, auth):
        """Wrap body in a soap:Enveloppe."""
        env = Environment(
            loader=PackageLoader('roulier', '/carriers/dpd/templates'),
            extensions=['jinja2.ext.with_'])

        template = env.get_template("dpd_soap.xml")
        body_stripped = remove_empty_tags(body)
        header_template = env.get_template("dpd_header.xml")
        header_xml = header_template.render(auth=auth)
        data = template.render(body=body_stripped, header=header_xml)
        return data.encode('utf8')

    def send_request(self, body):
        """Send body to dpd WS."""
        return requests.post(
            self.DPD_WS,
            headers={'content-type': 'text/xml'},
            data=body)

    def handle_500(self, response):
        """Handle reponse in case of ERROR 500 type."""
        log.warning('Dpd error 500')
        obj = objectify.fromstring(response.content)
        errors = [{
            "id": obj.xpath('//faultcode')[0],
            "message": obj.xpath('//faultstring')[0],
        }]
        raise CarrierError(response, errors)

    def handle_200(self, response):
        """Handle response type 200 (success)."""
        def extract_soap(response_xml):
            obj = objectify.fromstring(response_xml)
            return obj.Body.getchildren()[0]

        body = extract_soap(response.content)
        body_xml = etree.tostring(body)
        return {
            "body": body_xml,
            "response": response,
        }

    def handle_response(self, response):
        """Handle response of webservice."""
        if response.status_code == 200:
            return self.handle_200(response)
        elif response.status_code == 500:
            return self.handle_500(response)
        else:
            raise CarrierError(response, [{
                'id': None,
                'message': "Unexpected status code from server",
            }])
